/*
  assumes zero index positions

  should remove mutexs, and make function thread safe. dragon daemon

  from .448 this is modified such that each fasta is wrapped within a struct
  we also assume that each chunk is from the same chr so we won't need to strcmp every site.

  There could be some speed penality when we change chromosomes.
  since we might clear/reload the fasta for the 2 chromosomes. But this code will work.

  Everytime we want to read a fastasequence a mutex is locked. This might be too slow, and might need modification.
  DRAGON
*/



#include "analysisFunction.h"
#include "shared.h"

#include "general.h"
#include "pthread.h"
#include "getFasta.h"

//this will initialize our data
perFasta *init(const char *fname){
  perFasta *r= new perFasta;
  r->fai = NULL;
  r->seqs = NULL;
  r->curChr = NULL;
  if(pthread_mutex_init(&r->aMut,NULL)){
    fprintf(stderr,"[%s:%s] error reading fai file:%s\n",__FILE__,__FUNCTION__,fname);
    exit(0);
  }

  if(NULL==(r->fai = fai_load(fname))){
    fprintf(stderr,"[%s:%s] error reading fai file:%s\n",__FILE__,__FUNCTION__,fname);
    exit(0);
  }
  
  return r;
}

//this will destroy a perfasta structure
void destroy(perFasta *f){
  fai_destroy(f->fai);
  free(f->seqs);
  free(f->curChr);
  f=NULL;
}


//this will load a chromome
char *loadChr(perFasta *f,const char*fname){
  //  fprintf(stderr,"\t->loading chr:%s from faidx=%p curchr=%s\n",fname,f,f->curChr);
  free(f->curChr);
  free(f->seqs);
  f->curChr = strdup(fname);
  f->seqs = faidx_fetch_seq(f->fai, f->curChr, 0, 0x7fffffff, &f->chrLen);  
  return f->seqs;
}



char *getFasta::magic(const loci *locs,int numSites,perFasta *f){
  pthread_mutex_lock(&f->aMut);

  //load new chr if different from last or if nothing has been loaded before
  if(f->curChr==NULL||strcmp(locs[0].chromo,f->curChr))
    loadChr(f,locs[0].chromo);

  //first check that last position is not to long after.
  if(locs[numSites-1].position>f->chrLen+200){
    fprintf(stderr,"Trying to access fasta efter end of chromsome+200:%s/%s pos=%d ref_len=%d\n",f->curChr,locs[0].chromo,locs[numSites-1].position,f->chrLen);
    fflush(stderr);
    exit(0);
  }
  //now loop over all positions. We allow +200 since a read might match to the end of a chr
  char *ret=new char[numSites];
    
  for(int i=0;i<numSites;i++){
    if(f->seqs&&locs[i].position<f->chrLen)
      ret[i] = refToInt[f->seqs[locs[i].position]];//offseting by one, to make in work like samtools
    else
      ret[i] = 4;
  }  
  return ret;
  pthread_mutex_lock(&f->aMut);
}



void getFasta::printArg(FILE *argFile,const char *refName,const char *ancName){
  fprintf(argFile,"%s:\n\n",__FILE__);
  fprintf(argFile,"-ref\t%s\n",refName);
  fprintf(argFile,"-anc\t%s\n",ancName);
  fprintf(argFile,"\n");

}

getFasta::~getFasta(){
  if(ref)
    destroy(ref);
  if(anc)
    destroy(anc);
}


void getFasta::getOptions(argStruct *arguments){
  //default
  char *refName = NULL;
  char *ancName = NULL;
  //from command line
  ancName = angsd::getArg("-anc",ancName,arguments);
  refName = angsd::getArg("-ref",refName,arguments);
  //  fprintf(stderr,"refName=%s\tancName=%s\n",refName,ancName);


  if(ancName!=NULL)
    if(strcmp(ancName,"-999")==0){
      ancName=NULL;
      printArg(stdout,refName,ancName);
      exit(0);
 
    }
  if(refName!=NULL)
    if(strcmp(refName,"-999")==0){
      refName=NULL;
      printArg(stdout,refName,ancName);
      exit(0);
 
    }


  if(ancName!=NULL)
    anc = init(ancName);

  if(refName!=NULL)
    ref = init(refName);
  //  fprintf(stderr,"ref=%p\tref->curChr=%s\n",ref,ref->curChr);
  printArg(arguments->argumentFile,refName,ancName);
 
}

 getFasta::getFasta(argStruct *arguments){
   anc=NULL;
   ref=NULL;
   getOptions(arguments);
   //  fprintf(stderr,"ref=%p anc=%p\n",ref,anc);

}

void getFasta::run(funkyPars *pars){
  if(pars->numSites==0)
    return;
  if(ref)
    pars->ref = magic(pars->sites,pars->numSites,ref);
  if(anc)
    pars->anc = magic(pars->sites,pars->numSites,anc);
  
}


void getFasta::print(funkyPars *f){


}
void getFasta::clean(funkyPars *f){


}
