#include <stdint.h>
#include <ctype.h>
#include "bams.h"
#include "kstring.h"
#include "printer.h"
#include "general.h"
#include "getFasta.h"
static const char *bam_nt16_rev_table = "=ACMGRSVTWYHKDBN";

extern general **allMethods;



void printAuxBuffered(uint8_t *s, uint8_t *sStop,kstring_t &str ) {
  //  fprintf(stderr,"\ncomp:%p vs %p\n",s,sStop);
  
  while (s < sStop) {
    uint8_t type;
    kputc('\t', &str);kputc(s[0], &str);kputc(s[1], &str); kputc(':', &str); 
    //    fprintf(stderr,"\t%c%c:",s[0],s[1]);
    s += 2; type = *s; ++s;
    //    fprintf(stderr,"\ntype=%c\n",type);//,(char)*s);
    //    kputc('\t', &str); kputsn((char*)key, 2, &str); kputc(':', &str);
    if (type == 'A') { kputsn("A:", 2, &str); kputc(*s, &str); ++s; }
    else if (type == 'C') { kputsn("i:", 2, &str); kputw(*s, &str); ++s; }
    else if (type == 'c') { kputsn("i:", 2, &str); kputw(*(int8_t*)s, &str); ++s; }
    else if (type == 'S') { kputsn("i:", 2, &str); kputw(*(uint16_t*)s, &str); s += 2; }
    else if (type == 's') { kputsn("i:", 2, &str); kputw(*(int16_t*)s, &str); s += 2; }
    else if (type == 'I') { kputsn("i:", 2, &str); kputuw(*(uint32_t*)s, &str); s += 4; }
    else if (type == 'i') { kputsn("i:", 2, &str); kputw(*(int32_t*)s, &str); s += 4; }
    else if (type == 'f') { ksprintf(&str, "f:%g", *(float*)s); s += 4; }
    else if (type == 'd') { ksprintf(&str, "d:%lg", *(double*)s); s += 8; }
    else if (type == 'Z' || type == 'H') { kputc(type, &str); kputc(':', &str); while (*s) kputc(*s++, &str); ++s; }
    else if (type == 'B') {
      uint8_t sub_type = *(s++);
      int32_t n;
      memcpy(&n, s, 4);
      s += 4; // no point to the start of the array
      kputc(type, &str); kputc(':', &str); kputc(sub_type, &str); // write the typing
      for (int i = 0; i < n; ++i) {
	kputc(',', &str);
	if ('c' == sub_type || 'c' == sub_type) { kputw(*(int8_t*)s, &str); ++s; }
	else if ('C' == sub_type) { kputw(*(uint8_t*)s, &str); ++s; }
	else if ('s' == sub_type) { kputw(*(int16_t*)s, &str); s += 2; }
	else if ('S' == sub_type) { kputw(*(uint16_t*)s, &str); s += 2; }
	else if ('i' == sub_type) { kputw(*(int32_t*)s, &str); s += 4; }
	else if ('I' == sub_type) { kputuw(*(uint32_t*)s, &str); s += 4; }
	else if ('f' == sub_type) { ksprintf(&str, "%g", *(float*)s); s += 4; }
      }
    }
  }
  //  fprintf(stderr,"done\n");
}



extern int rute;
size_t cnt[4] = {0,0,0,0};
void printReadBuffered(aRead &rd,aHead *hd,kstring_t &str) {
  str.l = 0;
  
  getFasta *gf =(getFasta *) allMethods[1];
  static int curChr = -1;
  static char *theRef = NULL;
  if(bam_validate1(hd,rd)==0){
    fprintf(stderr,"problems validateing\n");
    exit(0);
  }
  kputsn((char *)rd.vDat,rd.l_qname-1,&str);kputc('\t', &str);
  kputw((int)rd.flag_nc>>16, &str); kputc('\t', &str); 
  
  if(rute&&(curChr==-1||curChr!=rd.refID)){
    theRef = gf->hidden2(hd->name[rd.refID]);
    curChr = rd.refID;
    fprintf(stderr,"reading fasta for chr=%s\n",hd->name[rd.refID]);
    fflush(stderr);
  }
  if(rd.refID==-1)//unmatched read
    kputc('*', &str);
  else
    kputs(hd->name[rd.refID] , &str);
  kputc('\t', &str); 
  
  kputw(rd.pos+1, &str);   kputc('\t', &str); 
  kputw(rd.mapQ, &str);kputc('\t', &str); 
  
  
  int nCigs = rd.nCig;
  
  if(nCigs==0)
    kputc('*', &str);// if no cigars
  else{
    for (int i = 0; i < nCigs; ++i) {//print cigars
      uint32_t *cigs =getCig(&rd);
      kputw(cigs[i]>>BAM_CIGAR_SHIFT, &str);
      kputc("MIDNSHP"[cigs[i]&BAM_CIGAR_MASK], &str);
    }
  }
  kputc('\t', &str); 
  
  if(rd.next_refID==-1)
    kputc('*', &str);// if no cigars     
  else if(rd.refID==rd.next_refID)
    kputc('=', &str);
  else
    kputs(hd->name[rd.next_refID] , &str);
  kputc('\t', &str); 
  
  kputw(rd.next_pos+1, &str);   kputc('\t', &str); 
  kputw(rd.tlen, &str);   kputc('\t', &str); 
  
  //  fprintf(stderr,"STRAND = %d \n",bam1_strand(&rd));
  
  //start seq

  if(rute==0){
    //don't perform any modification
    char *seq = (char *)getSeq(&rd);
    for(int i=0;i<rd.l_seq;i++)
      kputc(bam_nt16_rev_table[bam1_seqi(seq, i)], &str); 
  
  }else{
    int refPos = rd.pos;
    int seq_pos = 0;
    char *seq = (char *)getSeq(&rd);
    uint32_t *cigs = getCig(&rd);
    
    if(nCigs==0)//if no cigar print all
      for(int i=0;i<rd.l_seq;i++)
	kputc(bam_nt16_rev_table[bam1_seqi(seq, i)], &str); 
    

    for(int i=0;i<nCigs;i++) {      
      int opCode = cigs[i]&BAM_CIGAR_MASK; //what to do
      int opLen = cigs[i]>>BAM_CIGAR_SHIFT; //length of what to do
      
      if(opCode&BAM_CINS) {

	//we don't modify the position or the read with an insertion
	for(int i=seq_pos;i<seq_pos+opLen;i++)
	  kputc(bam_nt16_rev_table[bam1_seqi(seq, i)], &str); 
	seq_pos += opLen;
      }else if(opCode&BAM_CDEL) {//this is the deletion part
	refPos += opLen;
      }else if(opCode&BAM_CSOFT_CLIP){
	//soft clip don't modify anything just print
	for(int i=seq_pos;i<seq_pos + opLen;i++)
	  kputc(bam_nt16_rev_table[bam1_seqi(seq, i)], &str); 
	seq_pos +=opLen;
      
      }else if(opCode==BAM_CMATCH) {
	for(int i=0;i<opLen;i++){
	  char s = bam_nt16_rev_table[bam1_seqi(seq, seq_pos+i)];
	  char r = toupper(theRef[refPos+i]);
	  if(seq_pos+i<rute) {
	    if(bam1_strand(&rd) && s=='T'&&r=='C' ){
	      s='N';
	      cnt[0]++;
	    }else if(bam1_strand(&rd)==1 && s=='A'&&r=='G' ){
	      s='N';
	      cnt[1]++;
	    }
	  }else if (rd.l_seq-seq_pos-i<rute){
	    if(bam1_strand(&rd)==1 && s=='T'&&r=='C' ){
	      s='N';
	      cnt[2]++;
	    }  else if(bam1_strand(&rd)==0 && s=='A'&&r=='G' ){
	      s='N';
	      cnt[3]++;
	    }	    

	  }
	  kputc(s, &str); 
	    
	  //	  kputc(bam_nt16_rev_table[bam1_seqi(seq, seq_pos+i)], &str); 
	}
	refPos =+ opLen;
	seq_pos += opLen;
      }else
	printErr();//unknown CIGAR


    }
  }
  
  kputc('\t', &str); 
  
  char *quals =(char *)getQuals(&rd);
  for(int i=0;i<rd.l_seq;i++)
    kputc(quals[i]+33, &str); 
  
  //   int startAux= getAuxStart(&rd);
  //   uint8_t *s=rd.vDat+startAux;
  //int lensOfAux = rd.block_size-startAux;
  
  //below is taken directly from samtools,(not to steal, to preserve ordering etc, all credits go where credit is due)
  
  //   fprintf(stdout,"%s",str.s);
  //from aux start to the last memadrs in chunk
  printAuxBuffered(getAuxStart(&rd),rd.vDat+rd.block_size,str);
  
  
  kputc('\n', &str); 
}


void printReadBuffered_old(aRead &rd,aHead *hd,kstring_t &str) {
  str.l = 0;
   
   if(bam_validate1(hd,rd)==0){
     fprintf(stderr,"problems validateing\n");
     exit(0);
   }
   kputsn((char *)rd.vDat,rd.l_qname-1,&str);kputc('\t', &str);
   kputw((int)rd.flag_nc>>16, &str); kputc('\t', &str); 
   
   
   if(rd.refID==-1)//unmatched read
     kputc('*', &str);
   else
     kputs(hd->name[rd.refID] , &str);
   kputc('\t', &str); 
   
   kputw(rd.pos+1, &str);   kputc('\t', &str); 
   kputw(rd.mapQ, &str);kputc('\t', &str); 


   int nCigs = rd.nCig;

   if(nCigs==0)
     kputc('*', &str);// if no cigars
   else{
     for (int i = 0; i < nCigs; ++i) {//print cigars
       uint32_t *cigs =getCig(&rd);
       kputw(cigs[i]>>BAM_CIGAR_SHIFT, &str);
       kputc("MIDNSHP"[cigs[i]&BAM_CIGAR_MASK], &str);
     }
   }
   kputc('\t', &str); 
   
   if(rd.next_refID==-1)
     kputc('*', &str);// if no cigars     
   else if(rd.refID==rd.next_refID)
     kputc('=', &str);
   else
     kputs(hd->name[rd.next_refID] , &str);
   kputc('\t', &str); 

   kputw(rd.next_pos+1, &str);   kputc('\t', &str); 
   kputw(rd.tlen, &str);   kputc('\t', &str); 



   //start seq
   if(0){//this is normal rute
     char *seq = (char *)getSeq(&rd);
     for(int i=0;i<rd.l_seq;i++)
       kputc(bam_nt16_rev_table[bam1_seqi(seq, i)], &str); 
   }else{
     //this is modification rute



   }


   kputc('\t', &str); 

   char *quals =(char *)getQuals(&rd);
   for(int i=0;i<rd.l_seq;i++)
     kputc(quals[i]+33, &str); 
   
   //   int startAux= getAuxStart(&rd);
   //   uint8_t *s=rd.vDat+startAux;
   //int lensOfAux = rd.block_size-startAux;

   //below is taken directly from samtools,(not to steal, to preserve ordering etc, all credits go where credit is due)

   //   fprintf(stdout,"%s",str.s);
   //from aux start to the last memadrs in chunk
   printAuxBuffered(getAuxStart(&rd),rd.vDat+rd.block_size,str);

   
   kputc('\n', &str); 
}
