#include <cstdio>
#include <vector>
#include <cstring>
#include <sys/stat.h>
#include <pthread.h>
#include "mUpPile.h"

#include "shared.h"
#include "general.h" //<-new way, to add additional analysis to angsd
#include "analysisFunction.h" //<-included fancy parsing of argument

//utility class to keep track of chunk order when dumping results
#include "printRes.h"

int andersSux =0;
general **allMethods = NULL;

extern pthread_mutex_t mUpPile_mutex;

int maxThreads=1;//this value is given from the commandline -nThreads
pthread_t thread1;
pthread_cond_t cvMaxThread;
pthread_mutex_t counterMut;
int curRunning =0;
int chunkNumber =1;
int n_phenotypes;
printRes printer;



void init(argStruct *arguments){
  maxThreads=angsd::getArg("-nThreads",maxThreads,arguments);

  //these are  analysis that might be performed
  allMethods = extra(andersSux,arguments->outfiles,arguments->inputtype,arguments);
  pthread_mutex_init(&counterMut, NULL);
  pthread_cond_init (&cvMaxThread, NULL);
  
  
  //  fprintf(stderr,"[%s] \n",__FILE__);
  fprintf(stderr,"\t->Starting analysis\n");
}


void destroy(){
  fprintf(stderr,"\t-> Calling destroy\n");
  while(1){
    pthread_mutex_lock(&counterMut);
    if(curRunning==0)
      break;
    fprintf(stderr," Waiting for nThreads:%d\n",curRunning);
    pthread_mutex_unlock(&counterMut);
    sleep(1);
  }
  pthread_mutex_unlock(&counterMut);

  for(int i=0;i<andersSux;i++)
    delete allMethods[i];

  delete [] allMethods;

}


loci *populateLoci(const chunkyT *chk){
  loci *ret = new loci[chk->nSites];
  for(int i=0;i<chk->nSites;i++){
    ret[i].chromo = strdup(chk->refName);
    ret[i].position = chk->nd[i][0].refPos;
  }

  return ret;
}




void cleanUptNodeArray(tNode *row,int nSamples){
  //  fprintf(stderr,"[%s]\n",__FUNCTION__);
    for(int i=0;i< nSamples;i++) {
      if(row[i].l2!=0)
	//	fprintf(stderr,"l3\t%d\n",chk->nd[s][i].l2);
	for(int j=0;j<row[i].l2;j++){
	//	fprintf(stderr,"removing indels\n");
	dalloc_node(row[i].insert[j]);
      }
      free(row[i].insert);
      dalloc_node(row[i]);
    }
    delete [] row;
}


chunkyT *splitFucker3000T(nodePoolT *dn,int nFiles);//from mUppile.cpp;

int main_analysis(funkyPars *p) {
  //  fprintf(stderr,"p=%p\tp->callback=%p\n",p,p->for_callback);
  //first step is to make a chunk of data from the sample "uppiles"
  if(p->for_callback!=NULL) {
    fcb *f = p->for_callback;
    //fprintf(stderr,"picking up fcb in main_analysis\n");
    chunkyT *chk = splitFucker3000T(f->dn,f->nFiles);
    for(int i=0;0&&i<chk->nSites;i++)
      fprintf(stderr,"pos[%d]\t%d\n",i,chk->nd[i][0].refPos+1);


    chk->regStart = f->regStart;
    chk->regStop = f->regStop;
    chk->refName = f->refName;
    //    fprintf(stderr,"regstart=%d regstop=%d\n",chk->regStart,chk->regStop);
    if(chk->refPos[0]>f->regStop){
      fprintf(stderr,"Problems with regions\n (chkFirst=%d lastReg=%d)\n",chk->refPos[0],f->regStop);
      fflush(stderr);
      exit(0);
    }

    //regstop and regstart are the usersupplied startstop region, strip ends if needed
    if(1&&(chk->nd[0][0].refPos<chk->regStart||chk->nd[chk->nSites-1][0].refPos>chk->regStop)){
      //      fprintf(stderr,"[%s]",__FILE__);
      //fprintf(stderr,"\t->Stripping outof region parts of chunkID=%d nchunkNsites=%d\n",p->chunkNumber,chk->nSites);
      //      fprintf(stderr,"\t->chk=(%d,%d)\tregs=(%d,%d)\n",chk->nd[0][0].refPos,chk->nd[chk->nSites-1][0].refPos,chk->regStart,chk->regStop);
      fflush(stderr);

      int at=0;
      for(int i=0;i<chk->nSites;i++){//can be written faster
	if(chk->nd[i][0].refPos<chk->regStart)//should cleanup
	  cleanUptNodeArray(chk->nd[i],chk->nSamples);
	//	  continue;
	else if(chk->nd[i][0].refPos<chk->regStop)
	  chk->nd[at++] = chk->nd[i];
	else{//should cleanup
	  //	  fprintf(stderr,"\ncleaning up i=%d\n",i);
	  cleanUptNodeArray(chk->nd[i],chk->nSamples);
	  //break;
	}
      }

      chk->nSites = at;
      //      fprintf(stderr,"nsties=%d\n",chk->nSites);
      
    }
    //    fprintf(stderr,"nsties=%d\n",chk->nSites);
    //now chk contains the merged data
    p->chk = chk;
    //now make the position vector
    p->sites =(loci*) populateLoci(chk);
    p->numSites=chk->nSites;
  
    p->nInd = chk->nSamples;  
  }
  //  fprintf(stderr,"p->numSites=%d \n",p->numSites);

  //run all methods (ORDER is defined in general.cpp)

  for(int i=0;i<andersSux;i++){
    allMethods[i]->run(p);
    //    fprintf(stderr,"%d\n" ,p->keepSites[0]);
  }
  return 1;
}



void interface(funkyPars *p){
  if(p->killSig==1){
    //    fprintf(stderr,"chunkNumber=%d is the killSig\n",p->chunkNumber);
    //    return ;
  }

  if(p->killSig==0)
    main_analysis(p);//dont do analysis on the killsig chunk
  
  printer.prop(p);
}

void *slave(void *ptr){
  funkyPars *p = (funkyPars *) ptr;
  interface(p);

  //this might moved to the interface to allow more chunks with data (no waiting for print)
  pthread_mutex_lock( &counterMut );
  curRunning--;
  pthread_cond_signal(&cvMaxThread);
  pthread_mutex_unlock( &counterMut );

}


void *slave2(void *ptr){
  funkyPars *p = (funkyPars *) ptr;
  if(p->killSig==0)
    main_analysis(p);
  printFunky(p);
  pthread_mutex_unlock( &counterMut );
}

void master2(funkyPars *p){
  //  fprintf(stderr,"[%s] Number of threads running=%d\n",__FUNCTION__,curRunning);
  pthread_mutex_lock( &counterMut );
  int iret1 = pthread_create( &thread1, NULL, slave2, (void*) p);
  pthread_detach(thread1);
}



void master(funkyPars *p){
  //  fprintf(stderr,"[%s] Number of threads running=%d\n",__FUNCTION__,curRunning);
  pthread_mutex_lock( &counterMut );
  curRunning++;
  if(curRunning==maxThreads){
    //fprintf(stderr,"We are running max number of threads will wait for finishing threads: %d\n",curRunning);
    pthread_cond_wait(&cvMaxThread, &counterMut);
    //fprintf(stderr,"Done waiting for finishing threads:%d\n",curRunning);
  }pthread_mutex_unlock( &counterMut );
  int iret1 = pthread_create( &thread1, NULL, slave, (void*) p);
  pthread_detach(thread1);
}

/*
  This is the function that determines whether or not to start a new thread
*/
void selector(funkyPars *p){
  //  fprintf(stderr,"selector \n");
  if(p==NULL){
    p = allocFunkyPars();
    p->killSig = 1;
  }
  p->extras = new void*[andersSux];//funky
  p->chunkNumber = chunkNumber++;
  
  if(maxThreads==1)
    interface(p);
  else if(maxThreads==2)//simple serialization, allow one chunk to be computated
    master2(p);
  else
    master(p);
}


/*
  initialize all pointers to zero
 */
funkyPars *allocFunkyPars(){
  funkyPars *r = new funkyPars;

  r->sites = NULL;
  r->counts = NULL;
  r->likes = NULL;
  r->post = NULL;
  r->phat = NULL;

  r->major = NULL;
  r->minor = NULL;
  r->depth = NULL;

  r->results = NULL;
  r->ref = NULL;
  r->anc= NULL;
  r->keepSites =NULL;
  //allocate structures;

  r->results = new funkyPrint;
  r->results->freq = new funkyPrintFreq;
  r->results->asso = new funkyPrintAsso;
  //set nullpointers
  r->results->asso->stat=NULL;
  r->results->asso->keepInd=NULL;
  r->results->asso->freq=NULL;
  r->results->asso->lrt_snp=NULL;
  r->results->freq->pml=NULL;
  r->results->freq->pmlun= NULL;
  r->results->freq->pEM=NULL;
  r->results->freq->pEMun=NULL;
  r->results->freq->keepInd=NULL;

  r->chk = NULL;
  r->for_callback = NULL;
  r->killSig =0;
  return r;
}




//only called from this file and tglfs.cpp
void deallocFunkyPars(funkyPars *p) {
  //  fprintf(stderr,"[%s] cleaning funkypars\n",__FUNCTION__);  
  delete [] p->results->freq->keepInd;
  delete p->results->freq;
  delete p->results->asso;
  delete p->results;

  if(p->depth!=NULL){
    for(int i=0;i<p->numSites;i++)
      delete [] p->depth[i];
    delete [] p->depth;
  }

  if(p->sites!=NULL){
    for(int i=0;i<p->numSites;i++)
      free(p->sites[i].chromo);
    delete [] p->sites;
  }
  if(1){
    if(p->anc!=NULL)
      delete []  p->anc;
    if(p->ref!=NULL)
      delete [] p->ref;
  }

  //cleanup extra stuff
  void cleanUpChunkyT(chunkyT *chk);//<- from mUpPile.h
  if(p->chk!=NULL){
    //    fprintf(stderr,"cleanuing up chunkyT\n");
    cleanUpChunkyT(p->chk);
    p->chk = NULL;
  }

  if(p->for_callback!=NULL){
    fcb *f = p->for_callback;
    for(int i=0;i<f->nFiles;i++)
      if(f->dn[i].l!=0)//this is nescesarry...
	delete [] f->dn[i].nds;
    delete [] f->dn;
    delete f;
    f=NULL;
  }
  delete [] p->extras;
  delete p;

}
//plus one, plus 33
void printChunkyT(chunkyT *chk,double **liks,char *refs,FILE *fp){
  for(int s=0;s<chk->nSites;s++){
    fprintf(fp,"%s\t%d\t",chk->refName,chk->nd[s][0].refPos+1);
    if(refs!=NULL)
      fprintf(fp,"%c\t",intToRef[refs[s]]);
    for(int n=0;n<chk->nSamples;n++){
      tNode &nd = chk->nd[s][n];
      fprintf(fp,"%d\t",nd.l);
      for(int i=0;i<nd.l;i++)
	fprintf(fp,"%c",nd.seq[i]);
      fprintf(fp,"\t");
      for(int i=0;i<nd.l;i++)
	fprintf(fp,"%c",nd.qs[i]+33);
      fprintf(fp,"\t");
      for(int i=0;i<10;i++)
	fprintf(fp,"%f ",liks[s][n*10+i]);
    }
    fprintf(fp,"\n");
  }
}

//only one instance at a time is running this function
void printFunky(funkyPars *p){

  
  if(p->killSig==0){//don't print the empty killSig chunk
    if((p->chunkNumber%100)==0)
      fprintf(stderr,"\r\t->printing at chr: %s pos:%d chunknumber %d",p->sites[0].chromo,p->sites[0].position,p->chunkNumber);

    //    fprintf(stderr,"[%s] printing and cleaning: chunkNumber=%d\n",__FUNCTION__,p->chunkNumber);
    //printChunkyT(p->chk,p->likes,p->ref,fppp);// <- for debug

    //NB! the order is important!!!
    
    //  fprintf(stderr,"\n");
    for(int i=0;i<andersSux;i++){
      //      fprintf(stderr,"i=%d p=%p ",i,allMethods[i]);
      allMethods[i]->print(p);
      //      fprintf(stderr,"done\n");
    }
    for(int i=0;i<andersSux;i++)
      allMethods[i]->clean(p);


    //cleanup begin  
    if(p->major!=NULL)
      delete[] p->major;
    if(p->minor!=NULL)
      delete[] p->minor;

  delete[] p->keepSites;
    
    for(int i=0;i<0&&i<p->numSites;p++)
      free(p->sites[i].chromo);
  }else{
    pthread_mutex_unlock(&mUpPile_mutex);
    fprintf(stderr,"\n");//after last positions print add a neu line mutafuka

  }
  deallocFunkyPars(p);
}

//////////////////////////////////////////////////////////////////////////
//   Stuff to remove from this file
//////////////////////////////////////////////////////////////////////////


// a,c,g,t,n
// A,C,G,T,N
// 0,1,2,3,4
int refToInt[256] = {
  0,1,2,3,4,4,4,4,4,4,4,4,4,4,4,4,//15
  4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,//31
  4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,//47
  0,1,2,3,4,4,4,4,4,4,4,4,4,4,4,4,//63
  4,0,4,1,4,4,4,2,4,4,4,4,4,4,4,4,//79
  4,4,4,4,3,4,4,4,4,4,4,4,4,4,4,4,//95
  4,0,4,1,4,4,4,2,4,4,4,4,4,4,4,4,//111
  4,4,4,4,3,4,4,4,4,4,4,4,4,4,4,4,//127
  4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,//143
  4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,//159
  4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,//175
  4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,//191
  4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,//207
  4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,//223
  4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,//239
  4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4//255
};

char intToRef[5] = {'A','C','G','T','N'};



int fexists(const char* str){///@param str Filename given as a string.
  struct stat buffer ;
  return (stat(str, &buffer )==0 ); /// @return Function returns 1 if file exists.
}

size_t fsize(const char* fname){
  struct stat st ;
  stat(fname,&st);
  return st.st_size;
}

std::vector <char *> dumpedFiles;//small hack for getting a nice vector of outputfiles
FILE *openFile(const char* a,const char* b){
  if(0)
    fprintf(stderr,"[%s] %s %s",__FUNCTION__,a,b);
  char *c = new char[strlen(a)+strlen(b)+1];
  strcpy(c,a);
  strncat(c,b,strlen(b));
  //  fprintf(stderr,"\t-> Dumping file: %s\n",c);
  dumpedFiles.push_back(strdup(c));
  FILE *fp = fopen(c,"w");
  delete [] c;
  return fp;
}
gzFile openFileGz(const char* a,const char* b,const char *mode){
  if(0)
    fprintf(stderr,"[%s] %s%s\n",__FUNCTION__,a,b);
  char *c = new char[strlen(a)+strlen(b)+1];
  strcpy(c,a);
  strncat(c,b,strlen(b));
  //  fprintf(stderr,"\t-> Dumping file: %s\n",c);
  dumpedFiles.push_back(strdup(c));
  gzFile fp = getGz(c,mode);
  delete [] c;
  return fp;
}


FILE *getFILE(const char*fname,const char* mode){
  int writeFile = 0;
  for(size_t i=0;i<strlen(mode);i++)
    if(mode[i]=='w')
      writeFile = 1;
  FILE *fp;
  if(NULL==(fp=fopen(fname,mode))){
    fprintf(stderr,"\t->Error opening FILE handle for file:%s exiting\n",fname);
    exit(0);
  }
  return fp;
}

FILE *openFile(const char* a,const char* b,int i){
  char ary[5000];
  snprintf(ary,5000,"%s%d",b,i);
  return openFile(a,ary);  
}

void popBack(const char *a,const char*b,int to){
  char ary[5000];
  snprintf(ary,5000,"%s0-%d",b,to);
  dumpedFiles.push_back(strdup(ary));
}


gzFile getGz(const char*fname,const char* mode){
  int writeFile = 0;
  for(size_t i=0;i<strlen(mode);i++)
    if(mode[i]=='w')
      writeFile = 1;

  //  fprintf(stderr,"\t-> opening: %s\n",fname);
  gzFile fp=Z_NULL;
  if(NULL==(fp=gzopen(fname,mode))){
    fprintf(stderr,"\t-> Error opening gzFile handle for file:%s exiting\n",fname);
    exit(0);
  }
  return fp;
}
